#!/usr/bin/env python3

# Copyright: (c) 2020, Eugene Kovalev <eugene@kovalev.io>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from datetime import datetime, timedelta
from pathlib import Path
from uuid import uuid4 as uuid
from ansible.module_utils.basic import AnsibleModule
from cryptography.hazmat.primitives.hashes import SHA512
from cryptography.hazmat.primitives.serialization import NoEncryption
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PrivateFormat, load_pem_private_key
from cryptography.x509 import CertificateBuilder, Name, NameAttribute, BasicConstraints
from cryptography.x509 import KeyUsage, AuthorityKeyIdentifier, SubjectKeyIdentifier
from cryptography.x509 import load_pem_x509_certificate
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

DOCUMENTATION = '''
---
module: ca

short_description: An ansible module for managing a certificate authority. Pairs with the cert module.

version_added: "0.1"

description:
    - "This module can generate a CA key and certificate, as well as generate a new certificate for the same if the existing certificate is close to expiry. The module is designed with idempotency in mind, so it will only re-generate a certificate if one is close to expiry or if one of the properties does not match, implying that it has changed. This module will not create a new CA key if and only if one does not already exist. Even so, to prevent serious issues like loss of a CA certificate due to an accidental change, this module should be used in concert with version control."

options:
    ca_dir:
        description:
            - The directory where the certificate authority files will be kept. The CA cert and key will be at: <ca_dir>/<common_name>.crt and <ca_dir>/<common_name>.key respectively, while all generated keys and issued certificates will be at <ca_dir>/issued/<host_name>/ (assuming that complimenting private_key and certificate modules are used. If the directory does not exist, it will be created with the following permissions: u=rwx,g=o=
        required: True
    common_name:
        description:
            - This is the common name of the CA. It will be used as part of the file save-path: <common_name>.key and <common_name>.crt
        required: true
    country:
        description:
            - Country, part of the subject name
        required: true
    locality:
        description:
            - Locality, part of the subject name
        required: true
    state_or_province:
        description:
            - State/province, part of the subject name
        required: true
    org:
        description:
            - Organization name, part of the subject name
        required: true
    org_unit:
        description:
            - Organizational unit part of the subject name
        required: true
    email:
        description:
            - The email address associated with the CA
        required: true
    expire_days:
        description:
            - In the event that the module has to create a new certificate, this value will be used to determine when the cert should expire (measured in days).
        required: true
    renew_if_expiring_before:
        description:
            - If an existing certificate is expiring before today + the number of days provided in this value, the module will regenerate the certificate. Note that the expire_days value is less than renew_if_expiring_before, the certificate will be re-generated every time the module is used.
        required: true

author:
    - Eugene Kovalev (@cyclicircuit)
'''

EXAMPLES = '''
# Create a certificate authority on the ansible controller (can be used in concert with copy module to distribute the certificate
- name: Create a certificate authority called "Skynet" for 20 years with a renewal 30 days before expiry.
  ca:
    ca_dir: "{{ inventory_dir }}/host_files/pki/skynet/"
    common_name: skynet-ca
    country: US
    state_or_province: MA
    locality: Boston
    org: Kovalev
    org_unit: Skynet Network
    email: eugene@kovalev.io
    expire_days: 7300 # 20 years
    renew_if_expiring_before: 30
  delegate_to: 127.0.0.1    
'''

RETURN = '''
message:
    description: A message regarding why the module generated a new certificate if the previous certificate did not meet a constraint.
    type: str
    returned: always
ca_dir_created:
    description: Indicates whether the CA directory was created or not
    type: bool
    returned: always
key_created:
    description: Indicates whether the CA private key was created or not
    type: bool
    returned: always
'''

def new_private_key():
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend())

def name(common_name, country_name, locality_name, province_name,
         org_name, org_unit_name, email):
    return Name([
        NameAttribute(NameOID.COUNTRY_NAME, country_name),
        NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, province_name),
        NameAttribute(NameOID.LOCALITY_NAME, locality_name),
        NameAttribute(NameOID.ORGANIZATION_NAME, org_name),
        NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, org_unit_name),
        NameAttribute(NameOID.COMMON_NAME, common_name),
        NameAttribute(NameOID.EMAIL_ADDRESS, email)])

def new_ca_cert(common_name,
                country_name,
                locality_name,
                province_name,
                org_name,
                org_unit_name,
                email,
                expire_days,
                serial_number,
                ca_key):
    ca_subj_name = name(common_name, country_name, locality_name, province_name,
                        org_name, org_unit_name, email)
    ski_ext = SubjectKeyIdentifier.from_public_key(ca_key.public_key())
    builder = CertificateBuilder()
    builder = builder.subject_name(ca_subj_name)
    builder = builder.issuer_name(ca_subj_name)
    builder = builder.not_valid_before(datetime.today() - timedelta(days=1))
    builder = builder.not_valid_after(datetime.today() + timedelta(days=expire_days))
    builder = builder.serial_number(serial_number)
    builder = builder.public_key(ca_key.public_key())
    builder = builder.add_extension(BasicConstraints(ca=True, path_length=None), critical=False)
    builder = builder.add_extension(KeyUsage(digital_signature=False,
                                             content_commitment=False,
                                             key_encipherment=False,
                                             data_encipherment=False,
                                             key_agreement=False,
                                             key_cert_sign=True,
                                             crl_sign=True,
                                             encipher_only=False,
                                             decipher_only=False), critical=False)
    builder = builder.add_extension(AuthorityKeyIdentifier.from_issuer_subject_key_identifier(ski_ext), critical=False)
    builder = builder.add_extension(ski_ext, critical=False)
    return builder.sign(private_key=ca_key, algorithm=SHA512(), backend=default_backend())

def key_to_file(key, file_path):
    file_path.open('wb+').write(key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=NoEncryption()))

def cert_to_file(cert, file_path):
    file_path.open('wb+').write(cert.public_bytes(
        encoding=Encoding.PEM))

def key_from_file(file_path):
    return load_pem_private_key(file_path.open('rb').read(), None, default_backend())

def satisfies_constraints(cert_path, common_name, country_name, locality_name,
                          province_name, org_name, org_unit_name, email,
                          expiring_after_days):
    if not cert_path.exists():
        return False, "Certificate does not exist"
    cert = load_pem_x509_certificate(cert_path.open('rb').read(), default_backend())
    name_should_be = name(common_name, country_name, locality_name, province_name,
                           org_name, org_unit_name, email)
    expiring_soon = datetime.now() + timedelta(days=expiring_after_days) > cert.not_valid_after
    if expiring_soon:
        return False, f"The certificate {cert_path} is expiring in less than {expiring_after_days} days: {str(cert.not_valid_after)}"
    if cert.issuer != name_should_be:
        return False, f"The certificate issuer name {cert.issuer} does not match {name_should_be}"
    if cert.subject != name_should_be:
        return False, f"The certificate subject name {cert.subject} does not match {name_should_be}"
    return True, 'Certificate properties match constraints, no changes required'

def run_module():
    module_args = dict(
        ca_dir=dict(type='str', required=True),
        common_name=dict(type='str', required=True),
        country=dict(type='str', required=True),
        locality=dict(type='str', required=True),
        state_or_province=dict(type='str', required=True),
        org=dict(type='str', required=True),
        org_unit=dict(type='str', required=True),
        email=dict(type='str', required=True),
        expire_days=dict(type='int', required=True),
        renew_if_expiring_before=dict(type='int', required=True))
    result = {'changed': False, 'message': '', 'ca_dir_created': False, 'key_created': False}
    module = AnsibleModule(argument_spec=module_args, supports_check_mode=True)

    ca_dir = Path(module.params['ca_dir'])
    common_name = module.params['common_name']
    country = module.params['country']
    locality = module.params['locality']
    province = module.params['state_or_province']
    org = module.params['org']
    org_unit = module.params['org_unit']
    email = module.params['email']
    expire_days = module.params['expire_days']
    renew_if_expiring_before = module.params['renew_if_expiring_before']
    serial = int(uuid())
    ca_key_path = Path(ca_dir) / (common_name + '.key')
    ca_crt_path = Path(ca_dir) / (common_name + '.crt')
    cert_ok, msg = satisfies_constraints(ca_crt_path, common_name, country, locality,
                                         province, org, org_unit, email,
                                         renew_if_expiring_before)
    
    if module.check_mode:
        if not ca_dir.exists():
            result['changed'] = True
            result['message'] = 'The CA dir and all contents would be created'
            result['ca_dir_created'] = True
            result['key_created'] = True
        elif not ca_key_path.exists():
            result['changed'] = True
            result['message'] = 'The CA key and certificate would be created'
            result['ca_dir_created'] = False
            result['key_created'] = True
        elif not ca_cert_path.exists():
            result['changed'] = True
            result['message'] = 'The CA certificate would be created'
            result['ca_dir_created'] = False
            result['key_created'] = False
        elif not cert_ok:
            result['changed'] = True
            result['message'] = msg
            result['ca_dir_created'] = False
            result['key_created'] = False
        else:
            result['changed'] = False
            result['message'] = 'No changes required'
            result['ca_dir_created'] = False
            result['key_created'] = False
    else: # not in check-mode, so we do all this for real
        if not ca_dir.exists():
            result['changed'] = True
            ca_dir.mkdir(0o700, parents=True, exist_ok=True)
            result['ca_dir_created'] = False
        if not ca_key_path.exists():
            result['changed'] = True
            key = new_private_key()
            key_to_file(key, ca_key_path)
            result['key_created'] = True
            # if we create a new key, we always create a new cert
            cert_to_file(new_ca_cert(common_name, country, locality, province,
                                     org, org_unit, email, expire_days, serial, key),
                         ca_crt_path)
            result['message'] = 'Generated new key and cert'
        elif not cert_ok:
            result['changed'] = True
            key = key_from_file(ca_key_path)
            cert_to_file(new_ca_cert(common_name, country, locality, province,
                                     org, org_unit, email, expire_days, serial, key), ca_crt_path)
            result['message'] = 'A new CA certificate was generated because: ' + msg
    module.exit_json(**result)

def main():
    cert_file = Path('test-ca.crt')
    key_file = Path('test-ca.key')
    key = new_private_key()
    cert = new_ca_cert('test-common-name',
                       'US',
                       'MA',
                       'Boston',
                       'Kovalev',
                       'Skynet Network',
                       'eugene@kovalev.io',
                       90,
                       9918063603715682735,
                       key)
    key_to_file(key, key_file)
    cert_to_file(cert, cert_file)
    assert satisfies_constraints(cert_file,
                          'test-common-name',
                          'US',
                          'MA',
                          'Boston',
                          'Kovalev',
                          'Skynet Network',
                          'eugene@kovalev.io',
                          30,
                          9918063603715682735)

if __name__ == '__main__':
    run_module()



from pprint import pprint

from pytest import mark, raises
from robotnikmq import Message

from legion_utils import NotificationMsg, AlertMsg, Priority, InfoMsg, ActivityMsg
from legion_utils import WarningMsg, ErrorMsg, CriticalMsg, broadcast_alert_msg, priority_of
from legion_utils import valid_priority

from tests.integration.utils import vagrant_test, CONFIG, META_QUEUE, broadcast_receive, does_not_raise


@mark.parametrize('candidate, exp_priority, expectation', [
    ('warning', Priority.WARNING, does_not_raise()),
    ('error', Priority.ERROR, does_not_raise()),
    ('critical', Priority.CRITICAL, does_not_raise()),
    ('activity', Priority.ACTIVITY, does_not_raise()),
    ('information', Priority.INFO, does_not_raise()),
    ('warn', Priority.WARNING, does_not_raise()),
    ('err', Priority.ERROR, does_not_raise()),
    ('crit', Priority.CRITICAL, does_not_raise()),
    ('info', Priority.INFO, does_not_raise()),
    ('w', Priority.WARNING, does_not_raise()),
    ('e', Priority.ERROR, does_not_raise()),
    ('c', Priority.CRITICAL, does_not_raise()),
    ('a', Priority.ACTIVITY, does_not_raise()),
    ('i', Priority.INFO, does_not_raise()),
    ('not a priority', None, raises(ValueError)),
    (2, Priority.WARNING, does_not_raise()),
    (3, Priority.ERROR, does_not_raise()),
    (4, Priority.CRITICAL, does_not_raise()),
    (1, Priority.ACTIVITY, does_not_raise()),
    (0, Priority.INFO, does_not_raise()),
    (5, None, raises(ValueError)),
    (-1, None, raises(ValueError)),
    (Priority.WARNING, Priority.WARNING, does_not_raise()),
    (Priority.ERROR, Priority.ERROR, does_not_raise()),
    (Priority.CRITICAL, Priority.CRITICAL, does_not_raise()),
    (Priority.ACTIVITY, Priority.ACTIVITY, does_not_raise()),
    (Priority.INFO, Priority.INFO, does_not_raise()),
])
def test_priority_of(candidate, exp_priority, expectation):
    with expectation:
        assert priority_of(candidate) == exp_priority


@mark.parametrize('candidate, exp', [
    ('warning', True),
    ('error', True),
    ('critical', True),
    ('activity', True),
    ('information', True),
    ('warn', True),
    ('err', True),
    ('crit', True),
    ('info', True),
    ('w', True),
    ('e', True),
    ('c', True),
    ('a', True),
    ('i', True),
    ('not a priority', False),
    (2, True),
    (3, True),
    (4, True),
    (1, True),
    (0, True),
    (5, False),
    (-1, False),
    (None, False),
    (Priority.INFO, True),
    (Priority.ACTIVITY, True),
    (Priority.WARNING, True),
    (Priority.ERROR, True),
    (Priority.CRITICAL, True),
])
def test_valid_priority(candidate, exp):
    assert valid_priority(candidate) == exp


@mark.parametrize('contents,key,desc,ttl,prio,exp_contents,exp_key,exp_desc,exp_ttl,exp_prio,expectation',[
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30, Priority.INFO,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, Priority.INFO, does_not_raise()),
    ({'stuff': 'something'}, 'abc', 'stuff', 30, Priority.INFO,
     {'stuff': 'something'}, 'abc', 'stuff', 30, Priority.INFO, does_not_raise()),
    ({'stuff': 'something'}, None, None, None, None,
     {'stuff': 'something'}, None, None, None, Priority.INFO, does_not_raise()),
])
def test_notification_msg(contents, key, desc, ttl, prio,
                          exp_contents, exp_key, exp_desc, exp_ttl, exp_prio, expectation):
    with expectation:
        notification = NotificationMsg(contents, key, desc, ttl, prio)
        assert notification.contents == exp_contents
        assert notification.priority == exp_prio
        assert notification.ttl == exp_ttl
        assert notification.alert_key == exp_key
        assert notification.desc == exp_desc


@mark.parametrize('contents,ttl,exp_contents,exp_ttl,exp_prio,expectation', [
    ({'stuff': 'something'}, 30,
     {'stuff': 'something'}, 30, Priority.INFO, does_not_raise()),
    ({'stuff': 'something'}, 30,
     {'stuff': 'something'}, 30, Priority.INFO, does_not_raise()),
    ({'stuff': 'something'}, None,
     {'stuff': 'something'}, None, Priority.INFO, does_not_raise()),
])
def test_info_msg(contents, ttl, exp_contents, exp_ttl, exp_prio, expectation):
    with expectation:
        msg = InfoMsg(contents, ttl)
        assert msg.contents == exp_contents
        assert msg.priority == exp_prio
        assert msg.ttl == exp_ttl
        assert msg.alert_key is None
        assert msg.desc is None


@mark.parametrize('contents,ttl,exp_contents,exp_ttl,exp_prio,expectation', [
    ({'stuff': 'something'}, 30,
     {'stuff': 'something'}, 30, Priority.ACTIVITY, does_not_raise()),
    ({'stuff': 'something'}, 30,
     {'stuff': 'something'}, 30, Priority.ACTIVITY, does_not_raise()),
    ({'stuff': 'something'}, None,
     {'stuff': 'something'}, None, Priority.ACTIVITY, does_not_raise()),
])
def test_activity_msg(contents, ttl, exp_contents, exp_ttl, exp_prio, expectation):
    with expectation:
        msg = ActivityMsg(contents, ttl)
        assert msg.contents == exp_contents
        assert msg.priority == exp_prio
        assert msg.ttl == exp_ttl
        assert msg.alert_key is None
        assert msg.desc is None


@mark.parametrize('contents,key,desc,ttl,prio,exp_contents,exp_key,exp_desc,exp_ttl,exp_prio,expectation',[
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30, Priority.WARNING,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, Priority.WARNING, does_not_raise()),
    ({'stuff': 'something'}, 'abc', 'stuff', 30, Priority.WARNING,
     {'stuff': 'something'}, 'abc', 'stuff', 30, Priority.WARNING, does_not_raise()),
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', None, None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, Priority.WARNING, does_not_raise()),
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', None, Priority.INFO,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, Priority.WARNING, raises(ValueError)),
    ({'stuff': 'something'}, '', 'stuff', None, Priority.CRITICAL,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, Priority.CRITICAL, raises(ValueError)),
    ({'stuff': 'something'}, ['a', 'b', 'c'], '', None, Priority.CRITICAL,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, Priority.CRITICAL, raises(ValueError)),
])
def test_alert_msg(contents, key, desc, ttl, prio,
                   exp_contents, exp_key, exp_desc, exp_ttl, exp_prio, expectation):
    with expectation:
        alert = AlertMsg(contents, key, desc, ttl, prio)
        assert alert.contents == exp_contents
        assert alert.priority == exp_prio
        assert alert.ttl == exp_ttl
        assert alert.key == exp_key
        assert alert.description == exp_desc


def test_alert_msg_of():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': 60,
                   'description': 'Testing stuff',
                   'priority': Priority.ERROR,
                   'stuff': 'stuff'})
    alert = AlertMsg.of(msg)
    assert alert.contents['stuff'] == 'stuff'


def test_alert_msg_of_no_key():
    msg = Message({'ttl': 60,
                   'description': 'Testing stuff',
                   'priority': Priority.ERROR,
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


def test_alert_msg_of_no_description():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': 60,
                   'priority': Priority.ERROR,
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


def test_alert_msg_of_no_priority():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': 60,
                   'description': 'Testing stuff',
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


def test_alert_msg_of_no_ttl():
    msg = Message({'alert_key': '[stuff]',
                   'description': 'Testing stuff',
                   'priority': Priority.ERROR,
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


def test_alert_msg_of_invalid_priority():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': 60,
                   'description': 'Testing stuff',
                   'priority': 'Priority.ERROR',
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


def test_alert_msg_of_int_priority():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': 60,
                   'description': 'Testing stuff',
                   'priority': 3,
                   'stuff': 'stuff'})
    alert = AlertMsg.of(msg)
    assert alert.contents['stuff'] == 'stuff'
    assert alert.priority == Priority.ERROR


def test_alert_msg_of_invalid_ttl():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': '27',
                   'description': 'Testing stuff',
                   'priority': Priority.ERROR,
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


def test_alert_msg_of_negative_ttl():
    msg = Message({'alert_key': '[stuff]',
                   'ttl': -1,
                   'description': 'Testing stuff',
                   'priority': Priority.ERROR,
                   'stuff': 'stuff'})
    with raises(ValueError):
        AlertMsg.of(msg)
        assert False


@mark.parametrize('contents,key,desc,ttl,exp_contents,exp_key,exp_desc,exp_ttl,expectation',[
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, 'abc', 'stuff', 30,
     {'stuff': 'something'}, 'abc', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, '', 'stuff', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, raises(ValueError)),
    ({'stuff': 'something'}, ['a', 'b', 'c'], '', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, raises(ValueError)),
])
def test_warning_msg(contents, key, desc, ttl,
                     exp_contents, exp_key, exp_desc, exp_ttl, expectation):
    with expectation:
        alert = WarningMsg(contents, key, desc, ttl)
        assert alert.contents == exp_contents
        assert alert.priority == Priority.WARNING
        assert alert.ttl == exp_ttl
        assert alert.key == exp_key
        assert alert.description == exp_desc


@mark.parametrize('contents,key,desc,ttl,exp_contents,exp_key,exp_desc,exp_ttl,expectation',[
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, 'abc', 'stuff', 30,
     {'stuff': 'something'}, 'abc', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, '', 'stuff', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, raises(ValueError)),
    ({'stuff': 'something'}, ['a', 'b', 'c'], '', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, raises(ValueError)),
])
def test_error_msg(contents, key, desc, ttl,
                   exp_contents, exp_key, exp_desc, exp_ttl, expectation):
    with expectation:
        alert = ErrorMsg(contents, key, desc, ttl)
        assert alert.contents == exp_contents
        assert alert.priority == Priority.ERROR
        assert alert.ttl == exp_ttl
        assert alert.key == exp_key
        assert alert.description == exp_desc


@mark.parametrize('contents,key,desc,ttl,exp_contents,exp_key,exp_desc,exp_ttl,expectation',[
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, 'abc', 'stuff', 30,
     {'stuff': 'something'}, 'abc', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, does_not_raise()),
    ({'stuff': 'something'}, '', 'stuff', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, raises(ValueError)),
    ({'stuff': 'something'}, ['a', 'b', 'c'], '', None,
     {'stuff': 'something'}, '[a][b][c]', 'stuff', 30, raises(ValueError)),
])
def test_critical_msg(contents, key, desc, ttl,
                      exp_contents, exp_key, exp_desc, exp_ttl, expectation):
    with expectation:
        alert = CriticalMsg(contents, key, desc, ttl)
        assert alert.contents == exp_contents
        assert alert.priority == Priority.CRITICAL
        assert alert.ttl == exp_ttl
        assert alert.key == exp_key
        assert alert.description == exp_desc


@mark.parametrize('contents1,key1,desc1,ttl1,prio1,contents2,key2,desc2,ttl2,prio2,k_eq,d_eq,t_eq,p_eq,k_m,d_m', [
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30, Priority.WARNING,
     {'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30, Priority.WARNING,
     True, True, True, True, False, True),
    ({'stuff': 'something'}, ['a', 'b', 'c'], 'stuff', 30, Priority.WARNING,
     {'stuff': 'something1'}, ['a1', 'b1', 'c1'], 'stu1ff', 31, Priority.ERROR,
     False, False, False, False, False, False),
])
def test_alert_comparison(contents1, key1, desc1, ttl1, prio1,
                          contents2, key2, desc2, ttl2, prio2,
                          k_eq, d_eq, t_eq, p_eq, k_m, d_m):
    alert1 = AlertMsg(contents1, key1, desc1, ttl1, prio1)
    alert2 = AlertMsg(contents2, key2, desc2, ttl2, prio2)
    cmp = alert1.compare(alert2)
    assert cmp.key_equal == k_eq
    assert cmp.desc_equal == d_eq
    assert cmp.ttl_equal == t_eq
    assert cmp.priority_equal == p_eq
    assert cmp.key_match == k_m
    assert cmp.desc_match == d_m


@vagrant_test
def test_basic_broadcast_receive():
    def pub():
        NotificationMsg(contents={'stuff': 'Hello world!'},
                        priority=Priority.ACTIVITY, ttl=3).broadcast(META_QUEUE, 'test', CONFIG)

    def sub(msg):
        pprint(msg)

    broadcast_receive(pub, sub)


@vagrant_test
def test_basic_broadcast_alert():
    def pub():
        alert = ErrorMsg(contents={'stuff': 'Hello world!'}, key='stuff', desc='stuff', ttl=3)
        broadcast_alert_msg(META_QUEUE, 'test', alert, CONFIG)

    def sub(msg):
        pprint(msg)

    broadcast_receive(pub, sub)



from enum import IntEnum
from pprint import pformat
from re import match
from typing import Dict, Any, Optional, Union, List

from robotnikmq import Topic, Message, RobotnikConfig
from typeguard import typechecked


@typechecked
class Priority(IntEnum):
    INFO = 0
    ACTIVITY = 1
    WARNING = 2
    ERROR = 3
    CRITICAL = 4


INFO = {'info', 'information', 'i'}
ACTIVITY = {'activity', 'a'}
WARNING = {'warning', 'warn', 'w'}
ERROR = {'error', 'err', 'e'}
CRITICAL = {'critical', 'crit', 'c'}


@typechecked
def valid_priority(candidate: Any) -> bool:
    if isinstance(candidate, Priority):
        return True
    if isinstance(candidate, str):
        return any(candidate.lower() in P for P in [INFO, ACTIVITY, WARNING, ERROR, CRITICAL])
    if isinstance(candidate, int):
        return any(candidate == i for i in Priority)
    return False


@typechecked
def priority_of(candidate: Union[str, int]) -> Priority:
    if isinstance(candidate, str):
        if candidate.lower() in INFO:
            return Priority.INFO
        if candidate.lower() in ACTIVITY:
            return Priority.ACTIVITY
        if candidate.lower() in WARNING:
            return Priority.WARNING
        if candidate.lower() in ERROR:
            return Priority.ERROR
        if candidate.lower() in CRITICAL:
            return Priority.CRITICAL
    if isinstance(candidate, int):
        return Priority(candidate)
    raise ValueError(f'"{candidate}" cannot be mapped to a valid priority')


class NotificationMsg:
    @typechecked
    def __init__(self,
                 contents: Dict[str, Any],
                 alert_key: Union[str, List[str], None] = None,
                 desc: Optional[str] = None,
                 ttl: Optional[int] = None,
                 priority: Optional[Priority] = None):
        self.contents = contents
        self.alert_key = '['+']['.join(alert_key)+']' if isinstance(alert_key, List) else alert_key
        self.desc = desc
        self.ttl = ttl
        self.priority = priority or Priority.INFO

    @typechecked
    def broadcast(self, exchange: str, route: str, config: Optional[RobotnikConfig] = None) -> None:
        broadcast_msg(exchange, route, self, config)


@typechecked
class InfoMsg(NotificationMsg):
    def __init__(self,
                 contents: Dict[str, Any],
                 ttl: Optional[int] = None):
        super().__init__(contents=contents, ttl=ttl, priority=Priority.INFO)


@typechecked
class ActivityMsg(NotificationMsg):
    def __init__(self,
                 contents: Dict[str, Any],
                 ttl: Optional[int] = None):
        super().__init__(contents=contents, ttl=ttl, priority=Priority.ACTIVITY)


class AlertComparison:
    @typechecked
    def __init__(self, first: 'AlertMsg', second: 'AlertMsg'):
        self.key = first.key, second.key
        self.desc = first.desc, second.desc
        self.ttl = first.ttl, second.ttl
        self.priority = first.priority, second.priority
        self.contents = first.contents, second.contents

    @property
    def key_equal(self) -> bool:
        return self.key[0] == self.key[1]

    @property
    def desc_equal(self) -> bool:
        return self.desc[0] == self.desc[1]

    @property
    def ttl_equal(self) -> bool:
        return self.ttl[0] == self.ttl[1]

    @property
    def priority_equal(self) -> bool:
        return self.priority[0] == self.priority[1]

    @property
    def key_match(self) -> bool:
        return bool(match(*self.key) or match(self.key[1], self.key[0]))

    @property
    def desc_match(self) -> bool:
        return bool(match(*self.desc) or match(self.desc[1], self.desc[0]))


@typechecked
class AlertMsg(NotificationMsg):
    def __init__(self,
                 contents: Dict[str, Any],
                 key: Union[str, List[str]],
                 desc: str,
                 ttl: Optional[int] = None,
                 priority: Optional[Priority] = None,
                 from_msg: Optional[Message] = None):
        if priority is not None and priority < Priority.WARNING:
            raise ValueError('Alerts can only have a priority of WARNING (2) or higher')
        if not desc:
            raise ValueError('Alerts have to have a description')
        if not key:
            raise ValueError('Alerts have to have a key')
        super().__init__(desc=desc, priority=priority or Priority.WARNING,
                         ttl=ttl or 30, contents=contents)
        self.alert_key: str = '['+']['.join(key)+']' if isinstance(key, List) else key
        self.msg = from_msg

    def compare(self, other: 'AlertMsg') -> AlertComparison:
        return AlertComparison(self, other)

    @property
    def key(self) -> str:
        return self.alert_key

    @property
    def description(self) -> Optional[str]:
        return self.desc

    @staticmethod
    def of(msg: Message) -> 'AlertMsg':  # pylint: disable=C0103
        if 'ttl' not in msg.contents:
            raise ValueError(f'Message id {msg.msg_id} cannot be interpreted as an alert '
                             f'as it does not have a TTL: {pformat(msg.to_dict())}')
        if 'priority' not in msg.contents:
            raise ValueError(f'Message id {msg.msg_id} cannot be interpreted as an alert '
                             f'as it does not have a priority: {pformat(msg.to_dict())}')
        if not valid_priority(msg.contents['priority']):
            raise ValueError(f'Message id {msg.msg_id} cannot be interpreted as an alert '
                             f'as it does not have a valid priority: {pformat(msg.to_dict())}')
        if not isinstance(msg.contents['ttl'], int) or msg.contents['ttl'] < 0:
            raise ValueError(f'Message id {msg.msg_id} cannot be interpreted as an alert '
                             f'as it does not have a valid TTL: {pformat(msg.to_dict())}')
        if 'alert_key' not in msg.contents:
            raise ValueError(f'Message id {msg.msg_id} cannot be interpreted as an alert '
                             f'as it does not have an alert key: {pformat(msg.to_dict())}')
        if 'description' not in msg.contents:
            raise ValueError(f'Message id {msg.msg_id} cannot be interpreted as an alert '
                             f'as it does not have a description: {pformat(msg.to_dict())}')
        return AlertMsg(msg.contents, key=msg.contents['alert_key'],
                        desc=msg.contents['description'], ttl=msg.contents['ttl'],
                        priority=priority_of(msg.contents['priority']), from_msg=msg)


class WarningMsg(AlertMsg):
    @typechecked
    def __init__(self,
                 contents: Dict[str, Any],
                 key: Union[str, List[str]],
                 desc: str,
                 ttl: Optional[int] = None):
        if not desc:
            raise ValueError('Warnings (alerts) have to have a description')
        if not key:
            raise ValueError('Warnings (alerts) have to have a key')
        super().__init__(key=key, desc=desc, priority=Priority.WARNING,
                         ttl=ttl or 30, contents=contents)


class ErrorMsg(AlertMsg):
    @typechecked
    def __init__(self,
                 contents: Dict[str, Any],
                 key: Union[str, List[str]],
                 desc: str,
                 ttl: Optional[int] = None):
        if not desc:
            raise ValueError('Errors (alerts) have to have a description')
        if not key:
            raise ValueError('Errors (alerts) have to have a key')
        super().__init__(key=key, desc=desc, priority=Priority.ERROR,
                         ttl=ttl or 30, contents=contents)


class CriticalMsg(AlertMsg):
    @typechecked
    def __init__(self,
                 contents: Dict[str, Any],
                 key: Union[str, List[str]],
                 desc: str,
                 ttl: Optional[int] = None):
        if not desc:
            raise ValueError('Critical Alerts have to have a description')
        if not key:
            raise ValueError('Critical Alerts have to have a key')
        super().__init__(key=key, desc=desc, priority=Priority.CRITICAL,
                         ttl=ttl or 30, contents=contents)


@typechecked
def broadcast_msg(exchange: str, route: str, notification: NotificationMsg,
                  config: Optional[RobotnikConfig] = None) -> None:
    broadcast(exchange=exchange,
              route=route,
              priority=notification.priority,
              contents=notification.contents,
              ttl=notification.ttl,
              description=notification.desc,
              alert_key=notification.alert_key,
              config=config)


@typechecked
def broadcast_alert_msg(exchange: str, route: str, alert: AlertMsg,
                        config: Optional[RobotnikConfig] = None) -> None:
    broadcast(exchange=exchange,
              route=route,
              priority=alert.priority,
              contents=alert.contents,
              ttl=alert.ttl,
              description=alert.desc,
              alert_key=alert.alert_key,
              config=config)


@typechecked
def broadcast(exchange: str,
              route: str,
              priority: Priority,
              contents: Dict[str, Any],
              ttl: Optional[int] = None,
              description: Optional[str] = None,
              alert_key: Optional[str] = None,
              config: Optional[RobotnikConfig] = None):
    _contents: Dict[str, Any] = {'priority': priority.value}
    if priority.value >= 2:
        assert description is not None, 'Alerts (e.g. WARNING, ERROR, CRITICAL) must have a description'
        assert ttl is not None, 'Alerts (e.g. WARNING, ERROR, CRITICAL) must have a ttl (to clear an alert, set the ttl to 0)'
        assert alert_key is not None, 'Alerts (e.g. WARNING, ERROR, CRITICAL) must have an alert_key'
    if ttl is not None:
        _contents['ttl'] = ttl
    if description is not None:
        _contents['description'] = description
    if alert_key is not None:
        _contents['alert_key'] = alert_key
    _contents.update(contents)
    route += f'.{priority.name.lower()}'
    Topic(exchange=exchange, config=config).broadcast(Message(contents=_contents),
                                                      routing_key=route)


@typechecked
def broadcast_info(exchange: str, route: str, contents: Dict[str, Any],
                   config: Optional[RobotnikConfig] = None):
    broadcast(exchange, route, priority=Priority.INFO, contents=contents, config=config)


@typechecked
def broadcast_activity(exchange: str, route: str, contents: Dict[str, Any],
                       config: Optional[RobotnikConfig] = None):
    broadcast(exchange, route, priority=Priority.ACTIVITY, contents=contents, config=config)


@typechecked
def broadcast_alert(exchange: str,
                    route: str,
                    description: str,
                    alert_key: str,
                    contents: Dict[str, Any],
                    ttl: int = 30,
                    priority: Priority = Priority.WARNING,
                    config: Optional[RobotnikConfig] = None) -> None:
    broadcast(exchange, route, ttl=ttl, priority=priority, contents=contents,
              config=config, description=description, alert_key=alert_key)


@typechecked
def broadcast_warning(exchange: str, route: str, desc: str, alert_key: str,
                      contents: Dict[str, Any], ttl: int = 30,
                      config: Optional[RobotnikConfig] = None):
    broadcast_alert(exchange, route, description=desc, alert_key=alert_key, contents=contents,
                    ttl=ttl, priority=Priority.WARNING, config=config)


@typechecked
def broadcast_error(exchange: str, route: str, desc: str, alert_key: str,
                    contents: Dict[str, Any], ttl: int = 30,
                    config: Optional[RobotnikConfig] = None):
    broadcast_alert(exchange, route, description=desc, alert_key=alert_key, contents=contents,
                    ttl=ttl, priority=Priority.ERROR, config=config)


@typechecked
def broadcast_critical(exchange: str, route: str, desc: str, alert_key: str,
                       contents: Dict[str, Any], ttl: int = 30,
                       config: Optional[RobotnikConfig] = None):
    broadcast_alert(exchange, route, description=desc, alert_key=alert_key, contents=contents,
                    ttl=ttl, priority=Priority.CRITICAL, config=config)



from .core import (NotificationMsg,
                   InfoMsg,
                   ActivityMsg,
                   AlertMsg,
                   WarningMsg,
                   ErrorMsg,
                   CriticalMsg,
                   Priority,
                   broadcast,
                   broadcast_msg,
                   broadcast_alert_msg,
                   broadcast_info,
                   broadcast_activity,
                   broadcast_alert,
                   broadcast_warning,
                   broadcast_error,
                   broadcast_critical,
                   priority_of,
                   valid_priority)
from robotnikmq.log import log



#!/usr/bin/env python3

# Copyright: (c) 2020, Eugene Kovalev <eugene@kovalev.io>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from datetime import datetime, timedelta
from uuid import uuid4 as uuid
from pathlib import Path
from ipaddress import IPv4Address
from ansible.module_utils.basic import AnsibleModule
from cryptography.hazmat.primitives.hashes import SHA512
from cryptography.hazmat.primitives.serialization import NoEncryption
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PrivateFormat, load_pem_private_key
from cryptography.x509 import CertificateBuilder, Name, NameAttribute, BasicConstraints
from cryptography.x509 import KeyUsage, AuthorityKeyIdentifier, SubjectKeyIdentifier
from cryptography.x509 import load_pem_x509_certificate, ExtendedKeyUsage
from cryptography.x509 import SubjectAlternativeName, DNSName, IPAddress
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID, ExtensionOID

EXTENDED_KEY_USAGE_SERVER = [ExtendedKeyUsageOID.SERVER_AUTH,
                             ExtendedKeyUsageOID.CLIENT_AUTH,
                             ExtendedKeyUsageOID.CODE_SIGNING]

EXAMPLES = '''
# Create a certificate and private key with given DNS and IP SANs and sign it with a CA
- name: Create a certificate for a server called grab-e for 60 days with a renewal 30 days before expiry.
  cert:
    ca_dir: "{{ inventory_dir }}/host_files/pki/skynet/"
    ca: skynet-ca
    dns_sans: ['localhost', '']
    ip_sans: ['192.168.86.54', '127.0.0.1']
    common_name: grab-e
    country: US
    state_or_province: MA
    locality: Boston
    org: Kovalev
    org_unit: Skynet Network
    email: eugene@kovalev.io
    expire_days: 60
    renew_if_expiring_before: 30
  delegate_to: 127.0.0.1
'''

RETURN = '''
message:
    description: A message regarding why the module generated a new certificate if the previous certificate did not meet a constraint.
    type: str
    returned: always
ca_dir_created:
    description: Indicates whether the CA directory was created or not
    type: bool
    returned: always
key_created:
    description: Indicates whether the CA private key was created or not
    type: bool
    returned: always
'''

def new_private_key():
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend())

def key_to_file(key, file_path):
    file_path.open('wb+').write(key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=NoEncryption()))

def cert_to_file(cert, file_path):
    file_path.open('wb+').write(cert.public_bytes(
        encoding=Encoding.PEM))

def key_from_file(file_path):
    return load_pem_private_key(file_path.open('rb').read(), None, default_backend())

def cert_from_file(file_path):
    return load_pem_x509_certificate(file_path.open('rb').read(), default_backend())

def sans(cert):
    for ext in cert.extensions:
        if ext.oid == ExtensionOID.SUBJECT_ALTERNATIVE_NAME:
            return {str(n.value) for n in ext.value}

def name(common_name, country_name, locality_name, province_name,
         org_name, org_unit_name, email):
    return Name([
        NameAttribute(NameOID.COUNTRY_NAME, country_name),
        NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, province_name),
        NameAttribute(NameOID.LOCALITY_NAME, locality_name),
        NameAttribute(NameOID.ORGANIZATION_NAME, org_name),
        NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, org_unit_name),
        NameAttribute(NameOID.COMMON_NAME, common_name),
        NameAttribute(NameOID.EMAIL_ADDRESS, email)])

def new_cert(subject_name,
             expire_days,
             serial_number,
             ca_key,
             ca_cert,
             priv_key,
             dns_sans,
             ip_sans):
    return CertificateBuilder().subject_name(
        subject_name
    ).issuer_name(
        ca_cert.subject
    ).add_extension(
        SubjectKeyIdentifier.from_public_key(
            priv_key.public_key()
        ),
        critical=False
    ).add_extension(
        AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
            SubjectKeyIdentifier.from_public_key(
                ca_key.public_key()
            )
        ),
        critical=False
    ).not_valid_before(
        datetime.today() - timedelta(days=1)
    ).not_valid_after(
        datetime.today() + timedelta(days=expire_days)
    ).serial_number(
        serial_number
    ).public_key(
        priv_key.public_key()
    ).add_extension(
        BasicConstraints(ca=False, path_length=None),
        critical=False
    ).add_extension(
        ExtendedKeyUsage(EXTENDED_KEY_USAGE_SERVER),
        critical=False
    ).add_extension(
        KeyUsage(
            digital_signature=True,
            content_commitment=False,
            key_encipherment=True,
            data_encipherment=False,
            key_agreement=False,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False
        ),
        critical=False
    ).add_extension(
        SubjectAlternativeName(
            [DNSName(h) for h in dns_sans] + [IPAddress(IPv4Address(ip)) for ip in ip_sans]
        ),
        critical=False
    ).sign(
        private_key=ca_key,
        algorithm=SHA512(),
        backend=default_backend()
    )

def satisfies_constraints(cert, subject_name, issuer_name,
                          expiring_after_days, ip_sans, dns_sans):
    expiring_soon = datetime.now() + timedelta(days=expiring_after_days) > cert.not_valid_after
    if expiring_soon:
        return False, f"The certificate {cert} is expiring in less than {expiring_after_days} days: {str(cert.not_valid_after)}"
    if cert.issuer != issuer_name:
        return False, f"The certificate issuer name {cert.issuer} does not match {issuer_name}"
    if cert.subject != subject_name:
        return False, f"The certificate subject name {cert.subject} does not match {subject_name}"
    if sans(cert) != set(list(dns_sans) + list(ip_sans)):
        return False, f"The certificates subject-alt names {sans(cert)} do not match {set(list(dns_sans) + list(ip_sans))}"
    return True, 'Certificate properties match constraints, no changes required'

def run_module():
    module_args = dict(
        ca_dir=dict(type='str', required=True),
        ca=dict(type='str', required=True),
        common_name=dict(type='str', required=True),
        country=dict(type='str', required=True),
        locality=dict(type='str', required=True),
        state_or_province=dict(type='str', required=True),
        org=dict(type='str', required=True),
        org_unit=dict(type='str', required=True),
        email=dict(type='str', required=True),
        expire_days=dict(type='int', required=True),
        renew_if_expiring_before=dict(type='int', required=True),
        dns_sans=dict(type='list', elements='str', required=False),
        ip_sans=dict(type='list', elements='str', required=False))
    result = {'changed': False, 'message': '', 'dir_created': False, 'key_created': False}
    module = AnsibleModule(argument_spec=module_args, supports_check_mode=True)

    ca = module.params['ca']
    ca_dir = Path(module.params['ca_dir'])
    common_name = module.params['common_name']
    country = module.params['country']
    locality = module.params['locality']
    province = module.params['state_or_province']
    org = module.params['org']
    org_unit = module.params['org_unit']
    email = module.params['email']
    expire_days = module.params['expire_days']
    renew_if_expiring_before = module.params['renew_if_expiring_before']
    dns_sans = set(module.params['dns_sans'])
    ip_sans = set(module.params['ip_sans'])
    ca_key_path = Path(ca_dir) / (ca + '.key')
    ca_crt_path = Path(ca_dir) / (ca + '.crt')

    if not ca_key_path.exists() or not ca_crt_path.exists():
        module.fail_json(msg=f'The CA {ca} does not have a configured key and cert in {ca_dir}', **result)
    
    ca_key = key_from_file(ca_key_path)
    ca_crt = cert_from_file(ca_crt_path)
    server_dir = Path(ca_dir) / 'issued' / common_name
    key_path = server_dir / (common_name + '.key')
    crt_path = server_dir/ (common_name + '.crt')
    serial = int(uuid())
    cert_ok, msg = satisfies_constraints(cert_from_file(crt_path),
                                         name(common_name, country, locality, province, org, org_unit, email),
                                         ca_crt.subject,
                                         renew_if_expiring_before,
                                         dns_sans,
                                         ip_sans) if crt_path.exists() else (False,
                                                                             f'Certificate {common_name} does not exist')
    
    if module.check_mode:
        if not server.exists():
            result['changed'] = True
            result['message'] = 'The server dir and all contents would be created'
            result['dir_created'] = True
            result['key_created'] = True
        elif not key_path.exists():
            result['changed'] = True
            result['message'] = 'The server key and certificate would be created'
            result['dir_created'] = False
            result['key_created'] = True
        elif not crt_path.exists():
            result['changed'] = True
            result['message'] = 'The server certificate would be created'
            result['dir_created'] = False
            result['key_created'] = False
        elif not cert_ok:
            result['changed'] = True
            result['message'] = msg
            result['dir_created'] = False
            result['key_created'] = False
        else:
            result['changed'] = False
            result['message'] = 'No changes required'
            result['dir_created'] = False
            result['key_created'] = False
    else: # not in check-mode, so we do all this for real
        server_name = name(common_name, country, locality, province, org, org_unit, email)
        if not server_dir.exists():
            result['changed'] = True
            server_dir.mkdir(0o700, parents=True, exist_ok=True)
            result['dir_created'] = False
        if not key_path.exists():
            result['changed'] = True
            key = new_private_key()
            key_to_file(key, key_path)
            result['key_created'] = True
            # if we create a new key, we always create a new cert
            cert_to_file(new_cert(server_name, expire_days, serial, ca_key, ca_crt, key, dns_sans, ip_sans),
                         crt_path)
            result['message'] = 'Generated new key and cert'
        elif not cert_ok:
            result['changed'] = True
            key = key_from_file(key_path)
            cert_to_file(new_cert(server_name, expire_days, serial, ca_key, ca_crt, key, dns_sans, ip_sans),
                         crt_path)
            result['message'] = 'A new certificate was generated because: ' + msg
    module.exit_json(**result)

def main():
    key = new_private_key()
    ca_key = key_from_file(Path('test-ca.key'))
    ca_crt = cert_from_file(Path('test-ca.crt'))
    server_name = name('test-server', 'US', 'Boston', 'MA', 'Kovalev', 'Skynet Network', 'eugene@kovalev.io')
    server_cert = new_cert(server_name,
                           60,
                           9918063603715682735,
                           ca_key,
                           ca_crt,
                           key,
                           ['localhost', 'grab-e', 'grab-e.lan'],
                           ['127.0.0.1', '192.168.86.54', '10.0.0.12'])
    key_to_file(key, Path('test-server.key'))
    cert_to_file(server_cert, Path('test-server.crt'))
    print(satisfies_constraints(cert_from_file(Path('test-server.crt')),
                                server_name,
                                ca_crt.subject,
                                30, 9918063603715682735,
                                ['localhost', 'grab-e', 'grab-e.lan'],
                                ['127.0.0.1', '192.168.86.54', '10.0.0.12']))

if __name__ == '__main__':
    run_module()



from pprint import pprint

from legion_utils import broadcast, Priority, broadcast_info, broadcast_activity
from legion_utils import broadcast_alert, broadcast_warning, broadcast_error, broadcast_critical

from tests.integration.utils import vagrant_test, CONFIG, META_QUEUE, broadcast_receive

try:
    from pytest_cov.embed import cleanup_on_sigterm
except ImportError:
    pass
else:
    cleanup_on_sigterm()


@vagrant_test
def test_basic_broadcast_receive():
    def pub():
        broadcast(META_QUEUE, route='legion-utils-vm.test', ttl=3,
                  priority=Priority.ACTIVITY, contents={'stuff': 'Hello world!'},
                  config=CONFIG)

    def sub(msg):
        pprint(msg)

    broadcast_receive(pub, sub)


@vagrant_test
def test_broadcast_info():
    def pub():
        broadcast_info(META_QUEUE, route='legion-utils-vm.test',
                       contents={'stuff': 'Hello world!'},
                       config=CONFIG)

    def sub(msg):
        assert 'ttl' not in msg.contents
        assert msg.contents['priority'] == 0
        assert msg.routing_key == 'legion-utils-vm.test.info'

    broadcast_receive(pub, sub)


@vagrant_test
def test_broadcast_activity():
    def pub():
        broadcast_activity(META_QUEUE, route='legion-utils-vm.test',
                           contents={'stuff': 'Hello world!'},
                           config=CONFIG)

    def sub(msg):
        assert 'ttl' not in msg.contents
        assert msg.contents['priority'] == 1
        assert msg.routing_key == 'legion-utils-vm.test.activity'

    broadcast_receive(pub, sub)


@vagrant_test
def test_broadcast_alert():
    def pub():
        broadcast_alert(META_QUEUE, route='legion-utils-vm.test',
                        description='Testing stuff', alert_key='legion-utils-vm.test',
                        contents={'stuff': 'Hello world!'},
                        config=CONFIG)

    def sub(msg):
        assert msg.contents['ttl'] == 30
        assert msg.contents['priority'] == 2
        assert msg.routing_key == 'legion-utils-vm.test.warning'

    broadcast_receive(pub, sub)


@vagrant_test
def test_broadcast_warning():
    def pub():
        broadcast_warning(META_QUEUE, route='legion-utils-vm.test',
                          desc='Testing stuff', alert_key='legion-utils-vm.test',
                          contents={'stuff': 'Hello world!'},
                          config=CONFIG)

    def sub(msg):
        assert msg.contents['ttl'] == 30
        assert msg.contents['priority'] == 2
        assert msg.routing_key == 'legion-utils-vm.test.warning'

    broadcast_receive(pub, sub)


@vagrant_test
def test_broadcast_error():
    def pub():
        broadcast_error(META_QUEUE, route='legion-utils-vm.test',
                        desc='Testing stuff', alert_key='legion-utils-vm.test',
                        contents={'stuff': 'Hello world!'},
                        config=CONFIG)

    def sub(msg):
        assert msg.contents['ttl'] == 30
        assert msg.contents['priority'] == 3
        assert msg.routing_key == 'legion-utils-vm.test.error'

    broadcast_receive(pub, sub)


@vagrant_test
def test_broadcast_critical():
    def pub():
        broadcast_critical(META_QUEUE, route='legion-utils-vm.test',
                           desc='Testing stuff', alert_key='legion-utils-vm.test',
                           contents={'stuff': 'Hello world!'},
                           config=CONFIG)

    def sub(msg):
        assert msg.contents['ttl'] == 30
        assert msg.contents['priority'] == 4
        assert msg.routing_key == 'legion-utils-vm.test.critical'

    broadcast_receive(pub, sub)



from contextlib import contextmanager
from multiprocessing import Process, Event
from os.path import realpath
from pathlib import Path
from re import compile as regex
from subprocess import check_output
from time import sleep
from typing import Callable, Any, Tuple, Optional

from pytest import mark
from robotnikmq.config import server_config, RobotnikConfig
from robotnikmq.core import Message
from robotnikmq.subscriber import Subscriber, ExchangeBinding
from typeguard import typechecked


VAGRANT_VM_NAME = 'legion-utils-vm'
VAGRANT_RUNNING_PAT = regex(VAGRANT_VM_NAME + r'\s+running \(libvirt\)')
VAGRANT_CWD = Path(realpath(__file__)).parent.parent.parent

HERE = Path(realpath(__file__)).parent
CA_CERT = HERE / 'vagrant' / 'pki' / 'robotnik-ca.crt'
USERNAME = 'legion'
PASSWORD = 'hackme'
VIRTUAL_HOST = '/legion'
CERT = HERE / 'vagrant' / 'pki' / 'issued' / 'legion-utils-vm' / 'legion-utils-vm.crt'
KEY = HERE / 'vagrant' / 'pki' / 'issued' / 'legion-utils-vm' / 'legion-utils-vm.key'
PORT = 5671
LOCALHOST = '127.0.0.1'
META_QUEUE = 'skynet.legion'
CONFIG = RobotnikConfig(tiers=[[server_config(LOCALHOST, PORT, USERNAME, PASSWORD,
                                VIRTUAL_HOST, CA_CERT, CERT, KEY)]])


@contextmanager
def sub_process(target: Callable, args: Optional[Tuple[Any, ...]] = None,
                name: Optional[str] = None,
                terminate: bool = True):
    proc = Process(target=target, args=args or (), name=name)
    proc.start()
    try:
        sleep(0.2)
        yield proc
    finally:
        if terminate:
            proc.terminate()
        proc.join()


@typechecked
def vagrant_up() -> bool:
    return bool(VAGRANT_RUNNING_PAT.search(
        check_output(['vagrant', 'status', VAGRANT_VM_NAME], cwd=VAGRANT_CWD).decode()))


vagrant_test = mark.skipif(not vagrant_up(), reason="rabbitmq-vm needs to be running")


@typechecked
def broadcast_receive(pub: Callable, sub: Callable):
    msg_received = Event()

    def callback(msg: Message) -> None:
        sub(msg)
        msg_received.set()

    def subscriber():
        subscriber = Subscriber([ExchangeBinding(META_QUEUE, '#')],
                                config=CONFIG,
                                on_conn_error=lambda a: print(f'Could not connect: {a.args}'))
        subscriber.run(callback)

    sub_proc = Process(target=subscriber)
    sub_proc.start()
    sleep(0.2)
    pub_proc = Process(target=pub)
    pub_proc.start()
    assert msg_received.wait(timeout=5)
    pub_proc.terminate()
    sub_proc.terminate()
    pub_proc.join()
    sub_proc.join()


@contextmanager
def does_not_raise():
    yield
